@@ -23,6 +23,7 @@ import java.io.FileOutputStream;
 import java.io.OutputStream;
 import java.nio.file.Files;
 import java.nio.file.Paths;
+import java.util.Collections;
 import java.util.Dictionary;
 import java.util.Hashtable;
 import java.util.concurrent.atomic.AtomicReference;
@@ -36,6 +37,7 @@ import org.osgi.framework.Bundle;
 import org.osgi.framework.BundleContext;
 import org.osgi.framework.ServiceReference;
 import org.osgi.service.cm.Configuration;
+import org.osgi.service.cm.Configuration.ConfigurationAttribute;
 import org.osgi.service.cm.ConfigurationAdmin;
 import org.osgi.service.cm.ConfigurationEvent;
 
@@ -87,20 +89,22 @@ public class ConfigInstallerTest extends TestCase {
         String pid = file.getName().substring(0, file.getName().indexOf(".config"));
 
         Capture<Dictionary<String, Object>> props = new Capture<>();
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                 .andReturn(null);
         EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, "?"))
                 .andReturn(mockConfiguration);
+        EasyMock.expect(mockConfiguration.getAttributes()).andReturn(Collections.emptySet()).times(2);
+        EasyMock.expect(mockConfiguration.getPid()).andReturn(pid);
         EasyMock.expect(mockConfiguration.getProperties())
                 .andReturn(null);
         EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
                 .andReturn(null)
                 .anyTimes();
-        EasyMock.expect(mockBundleContext.getBundle())
-                .andReturn(mockBundle);
-        mockConfiguration.update(EasyMock.capture(props));
+        EasyMock.expect(mockConfiguration.updateIfDifferent(EasyMock.capture(props))).andReturn(true);
         EasyMock.expectLastCall();
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
         ci.install(file);
@@ -120,20 +124,23 @@ public class ConfigInstallerTest extends TestCase {
         String pid = file.getName().substring(0, file.getName().indexOf(".config"));
 
         Capture<Dictionary<String, Object>> props = new Capture<>();
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                 .andReturn(null);
         EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, "?"))
                 .andReturn(mockConfiguration);
+        EasyMock.expect(mockConfiguration.getAttributes())
+            .andReturn(Collections.emptySet()).times(2);
         EasyMock.expect(mockConfiguration.getProperties())
                 .andReturn(null);
+        EasyMock.expect(mockConfiguration.getPid()).andReturn(pid);
         EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
                 .andReturn(null)
                 .anyTimes();
-        EasyMock.expect(mockBundleContext.getBundle())
-                .andReturn(mockBundle);
-        mockConfiguration.update(EasyMock.capture(props));
+        EasyMock.expect(mockConfiguration.updateIfDifferent(EasyMock.capture(props))).andReturn(true);
         EasyMock.expectLastCall();
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
         ci.install(file);
@@ -146,11 +153,13 @@ public class ConfigInstallerTest extends TestCase {
 
     public void testGetNewFactoryConfiguration() throws Exception
     {
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                     .andReturn(null);
-        EasyMock.expect(mockConfigurationAdmin.createFactoryConfiguration( "pid", "?" ))
+        EasyMock.expect(mockConfigurationAdmin.getFactoryConfiguration( "pid", "factoryPid", "?" ))
                     .andReturn(mockConfiguration);
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -162,11 +171,13 @@ public class ConfigInstallerTest extends TestCase {
 
     public void testGetExistentFactoryConfiguration() throws Exception
     {
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                         .andReturn(null);
-        EasyMock.expect(mockConfigurationAdmin.createFactoryConfiguration( "pid", "?" ))
+        EasyMock.expect(mockConfigurationAdmin.getFactoryConfiguration( "pid", "factoryPid", "?" ))
                         .andReturn(mockConfiguration);
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -178,11 +189,13 @@ public class ConfigInstallerTest extends TestCase {
 
     public void testGetExistentNoFactoryConfiguration() throws Exception
     {
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                         .andReturn(null);
         EasyMock.expect(mockConfigurationAdmin.getConfiguration( "pid", "?" ))
                         .andReturn(mockConfiguration);
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -195,13 +208,17 @@ public class ConfigInstallerTest extends TestCase {
     public void testDeleteConfig() throws Exception
     {
         mockConfiguration.delete();
-        EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.LOG_LEVEL)).andReturn(null);
+        EasyMock.expect(mockConfiguration.getPid()).andReturn("pid");
         EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
+        EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
+            .andReturn(null)
+            .anyTimes();
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                         .andReturn(null);
         EasyMock.expect(mockConfigurationAdmin.getConfiguration("pid", "?" ))
                         .andReturn(mockConfiguration);
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -220,6 +237,9 @@ public class ConfigInstallerTest extends TestCase {
 
         final Capture<Dictionary<String, Object>> props = new Capture<>();
 
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
+        EasyMock.expect(mockConfiguration.getAttributes()).andReturn(Collections.emptySet()).times(2);
         EasyMock.expect(mockConfiguration.getProperties())
                 .andReturn(null);
         EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
@@ -231,7 +251,7 @@ public class ConfigInstallerTest extends TestCase {
                 .andReturn(mockConfiguration);
 
         ServiceReference<ConfigurationAdmin> sr = EasyMock.createMock(ServiceReference.class);
-        mockConfiguration.update(EasyMock.capture(props));
+        EasyMock.expect(mockConfiguration.updateIfDifferent(EasyMock.capture(props))).andReturn(true);
         EasyMock.expectLastCall();
 
         EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, "?"))
@@ -244,9 +264,9 @@ public class ConfigInstallerTest extends TestCase {
                     }
                 });
         EasyMock.expect(mockConfiguration.getPid())
-                .andReturn(pid);
+                .andReturn(pid).times(2);
 
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, sr);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle, sr);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
         ci.install(file);
@@ -264,6 +284,8 @@ public class ConfigInstallerTest extends TestCase {
 
         Capture<Dictionary<String, Object>> props = new Capture<>();
 
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockConfiguration.getProperties())
                 .andReturn(null);
         EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
@@ -278,7 +300,7 @@ public class ConfigInstallerTest extends TestCase {
         mockConfiguration.update(EasyMock.capture(props));
         EasyMock.expectLastCall();
 
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, sr);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle, sr);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -298,6 +320,8 @@ public class ConfigInstallerTest extends TestCase {
         Dictionary<String, Object> props = new Hashtable<>();
         props.put(DirectoryWatcher.FILENAME, file.toURI().toString());
 
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockConfiguration.getProperties())
                 .andReturn(props);
         EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
@@ -314,7 +338,7 @@ public class ConfigInstallerTest extends TestCase {
         mockConfiguration.update(EasyMock.capture(propsCapture));
         EasyMock.expectLastCall();
 
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, sr);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle, sr);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -335,11 +359,14 @@ public class ConfigInstallerTest extends TestCase {
 
         ServiceReference<ConfigurationAdmin> sr = EasyMock.createMock(ServiceReference.class);
 
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
                 .andReturn(null)
                 .anyTimes();
 
         final Configuration cachingPersistenceConfiguration = EasyMock.createMock(Configuration.class);
+        EasyMock.expect(cachingPersistenceConfiguration.getAttributes()).andReturn(Collections.emptySet()).times(2);
 
         final Dictionary<String, Object> cachedProps = new Hashtable<String, Object>() {
             {
@@ -353,17 +380,18 @@ public class ConfigInstallerTest extends TestCase {
                 .anyTimes();
 
         EasyMock.expect(cachingPersistenceConfiguration.getPid())
-                .andReturn(pid);
+                .andReturn(pid).times(2);
 
         EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, "?"))
                 .andReturn(cachingPersistenceConfiguration)
                 .anyTimes();
 
-        EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid))
+        EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, null))
                 .andReturn(cachingPersistenceConfiguration)
                 .anyTimes();
 
         final Configuration newConfiguration = EasyMock.createMock(Configuration.class);
+        EasyMock.expect(newConfiguration.getAttributes()).andReturn(Collections.emptySet()).times(2);
 
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                 .andReturn(new Configuration[] { newConfiguration });
@@ -375,21 +403,21 @@ public class ConfigInstallerTest extends TestCase {
                 .andReturn(pid);
 
         final Capture<Dictionary<String, Object>> props = new Capture<>();
-        newConfiguration.update(EasyMock.capture(props));
+        EasyMock.expect(newConfiguration.updateIfDifferent(EasyMock.capture(props))).andReturn(true);
 
-        cachingPersistenceConfiguration.update(EasyMock.capture(props));
+        EasyMock.expect(cachingPersistenceConfiguration.updateIfDifferent(EasyMock.capture(props)));
         EasyMock.expectLastCall()
-                .andAnswer(new IAnswer<Object>() {
+                .andAnswer(new IAnswer<Boolean>() {
                     @Override
-                    public Object answer() throws Throwable {
+                    public Boolean answer() throws Throwable {
 
                         cachedProps.put("networkInterface", props.getValue().get("networkInterface"));
 
-                        return null;
+                        return Boolean.TRUE;
                    }
                }).anyTimes();
 
-        EasyMock.replay(mockBundleContext, mockConfigurationAdmin, cachingPersistenceConfiguration, newConfiguration, sr);
+        EasyMock.replay(mockBundleContext, mockConfigurationAdmin, mockBundle, cachingPersistenceConfiguration, newConfiguration, sr);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -406,11 +434,14 @@ public class ConfigInstallerTest extends TestCase {
 
     public void testSetConfiguration() throws Exception
     {
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.CONFIG_ENCODING)).andReturn(null);
         EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.LOG_DEFAULT)).andReturn(null);
         EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.LOG_LEVEL)).andReturn(null);
-        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
         EasyMock.expect(mockConfiguration.getProperties()).andReturn(new Hashtable<String, Object>());
+        EasyMock.expect(mockConfiguration.getAttributes()).andReturn(Collections.emptySet()).times(2);
+        EasyMock.expect(mockConfiguration.getPid()).andReturn("firstcfg");
         EasyMock.reportMatcher(new IArgumentMatcher()
         {
             public boolean matches( Object argument )
@@ -423,12 +454,12 @@ public class ConfigInstallerTest extends TestCase {
                 buffer.append("<Dictionary check: testkey present?>");
             }
         } );
-        mockConfiguration.update(new Hashtable<String, Object>());
+        EasyMock.expect(mockConfiguration.updateIfDifferent(new Hashtable<String, Object>())).andReturn(true);
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                         .andReturn(null);
         EasyMock.expect(mockConfigurationAdmin.getConfiguration("firstcfg", "?"))
                         .andReturn(mockConfiguration);
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -436,12 +467,14 @@ public class ConfigInstallerTest extends TestCase {
 
         EasyMock.verify(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
     }
-    
-    public void testShouldSaveConfig() 
+
+    public void testShouldSaveConfig() throws Exception
     {
         final AtomicReference<Boolean> disable = new AtomicReference<Boolean>();
         final AtomicReference<Boolean> enable = new AtomicReference<Boolean>();
-        
+
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
         EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.DISABLE_CONFIG_SAVE)).andAnswer(
                 new IAnswer<String>() {
                     public String answer() throws Throwable {
@@ -456,7 +489,7 @@ public class ConfigInstallerTest extends TestCase {
                     }
                 }
         ).anyTimes();
-        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext, mockBundle);
 
         ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
 
@@ -499,4 +532,82 @@ public class ConfigInstallerTest extends TestCase {
         EasyMock.verify(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
     }
 
+    public void testSetREAD_ONLYFromFilePermissions() throws Exception
+    {
+        final File file = File.createTempFile("test", ".config");
+        try (OutputStream os = new FileOutputStream(file)) {
+            os.write("networkInterface = \"wlp3s1\"\n".getBytes("UTF-8"));
+        }
+        file.setReadOnly();
+
+        String pid = file.getName().substring(0, file.getName().indexOf(".config"));
+
+        ServiceReference<ConfigurationAdmin> sr = EasyMock.createMock(ServiceReference.class);
+
+        EasyMock.expect(mockBundleContext.getBundle()).andReturn(mockBundle).anyTimes();
+        EasyMock.expect(mockBundle.loadClass(ConfigurationAttribute.class.getName())).andReturn((Class)ConfigurationAttribute.class).anyTimes();
+        EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
+                .andReturn(null)
+                .anyTimes();
+
+        final Configuration cachingPersistenceConfiguration = EasyMock.createMock(Configuration.class);
+        EasyMock.expect(cachingPersistenceConfiguration.getAttributes()).andReturn(
+            Collections.singleton(ConfigurationAttribute.READ_ONLY)).anyTimes();
+        cachingPersistenceConfiguration.addAttributes(new ConfigurationAttribute[] {ConfigurationAttribute.READ_ONLY});
+        cachingPersistenceConfiguration.removeAttributes(EasyMock.anyObject(ConfigurationAttribute[].class));
+        EasyMock.expectLastCall().times(2);
+
+        final Dictionary<String, Object> cachedProps = new Hashtable<String, Object>() {
+            {
+                put("networkInterface", "wlp3s0");
+                put(DirectoryWatcher.FILENAME, file.toURI().toString());
+            }
+        };
+
+        EasyMock.expect(cachingPersistenceConfiguration.getPid())
+                .andReturn(pid).times(4);
+        EasyMock.expect(cachingPersistenceConfiguration.getProperties())
+                .andReturn(cachedProps)
+                .anyTimes();
+
+        EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, "?"))
+                .andReturn(cachingPersistenceConfiguration)
+                .anyTimes();
+
+        EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, null))
+                .andReturn(cachingPersistenceConfiguration)
+                .anyTimes();
+
+        final Configuration newConfiguration = EasyMock.createMock(Configuration.class);
+        EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
+                .andReturn(new Configuration[] { newConfiguration }).times(2);
+        EasyMock.expect(newConfiguration.getPid())
+                .andReturn(pid).times(2);
+
+        final Capture<Dictionary<String, Object>> props = new Capture<>();
+        EasyMock.expect(cachingPersistenceConfiguration.updateIfDifferent(EasyMock.capture(props)));
+        EasyMock.expectLastCall()
+                .andAnswer(new IAnswer<Boolean>() {
+                    @Override
+                    public Boolean answer() throws Throwable {
+
+                        cachedProps.put("networkInterface", props.getValue().get("networkInterface"));
+
+                        return Boolean.TRUE;
+                   }
+               }).anyTimes();
+
+        EasyMock.replay(mockBundleContext, mockConfigurationAdmin, mockBundle, cachingPersistenceConfiguration, newConfiguration, sr);
+
+        ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
+
+        ci.install(file);
+
+        file.setWritable(true);
+
+        ci.install(file);
+
+        EasyMock.verify(mockBundleContext, mockConfigurationAdmin, mockBundle, cachingPersistenceConfiguration, newConfiguration, sr);
+    }
+
 }
